<?php
namespace Ep\View\Renderer;

use Ep\View\Helper\TwigBridge;
use Interop\Container\ContainerInterface;
use Twig_Environment;
use Twig_Extension_Debug;
use Twig_Function;
use Twig_Loader_Filesystem;
use Twig_LoaderInterface;
use Zend\ServiceManager\Exception\ServiceNotCreatedException;
use Zend\ServiceManager\Factory\FactoryInterface;
use Zend\View\HelperPluginManager;
use Zend\View\Renderer\RendererInterface;

/**
 * Stvara novu instancu TwigRenderera.
 *
 * @package Ep\View\Renderer
 * @author Elektronika Perhoč <web@ep.hr>
 * @link http://gitlab.ep.local/ep-framework/core
 */
class TwigRendererFactory implements FactoryInterface
{
    /**
     * Stvara renderer.
     *
     * @param  ContainerInterface $container
     * @param  string $requestedName
     * @param  null|array $options
     *
     * @return RendererInterface
     *
     * @throws \Psr\Container\ContainerExceptionInterface
     * @throws \Psr\Container\NotFoundExceptionInterface
     * @throws \Twig_Error_Loader
     */
    public function __invoke(ContainerInterface $container, $requestedName, array $options = null)
    {
        $config = $options ?? $container->get('config')['twig'] ?? [];

        if ($container->get('config')['app']['development'] ?? false) {
            $config['options']['debug'] = true;
            $config['options']['auto_reload'] = true;
        }

        $twig = $this->createTwig($container, $config);
        $renderer = new TwigRenderer($twig);

        /** @var HelperPluginManager $plugins */
        $plugins = $container->get(HelperPluginManager::class);
        $renderer->setHelperPluginManager($plugins);

        $twig->addFunction(new Twig_Function('*', $plugins->get(TwigBridge::class), [
            'is_safe' => ['all'],
        ]));

        return $renderer;
    }

    /**
     * Stvara novi TwigLoader.
     *
     * @param array $config
     *
     * @return Twig_LoaderInterface
     *
     * @throws \Twig_Error_Loader
     */
    private function createLoader(array $config): Twig_LoaderInterface
    {
        if (!isset($config['template_paths']) || empty($config['template_paths'])) {
            throw new ServiceNotCreatedException("Polje 'template_paths' je obavezno.");
        }

        if (!is_array($config['template_paths'])) {
            throw new ServiceNotCreatedException("Polje 'template_paths' mora biti array.");
        }

        $loader = new Twig_Loader_Filesystem();

        foreach ($config['template_paths'] as $namespace => $path) {
            if (!is_array($path) && !is_dir($path)) {
                throw new ServiceNotCreatedException(sprintf(
                    "Nepostojeća putanja '%s'.", $path
                ));
            }

            $namespace = is_numeric($namespace) ? Twig_Loader_Filesystem::MAIN_NAMESPACE : $namespace;

            if (is_array($path)) {
                foreach ($path as $p) {
                    if (!is_dir($p)) {
                        throw new ServiceNotCreatedException(sprintf(
                            "Nepostojeća putanja '%s'.", $path
                        ));
                    }

                    $loader->prependPath($p, $namespace);
                }
            } else {
                $loader->prependPath($path, $namespace);
            }
        }

        return $loader;
    }

    /**
     * @param ContainerInterface $container
     * @param array $config
     *
     * @return Twig_Environment
     *
     * @throws \Psr\Container\ContainerExceptionInterface
     * @throws \Psr\Container\NotFoundExceptionInterface
     * @throws \Twig_Error_Loader
     */
    private function createTwig(ContainerInterface $container, array $config): Twig_Environment
    {
        $loader = $this->createLoader($config);
        $twig = new Twig_Environment($loader, $config['options'] ?? []);

        foreach (array_unique($config['extensions'] ?? []) as $extension) {
            $twig->addExtension($container->get($extension));
        }

        if ($config['options']['debug'] ?? false) {
            $twig->addExtension(new Twig_Extension_Debug());
        }

        return $twig;
    }
}