<?php
namespace EpTest\Http\Middleware;

use Ep\Cache\AbstractCacheFactory;
use Ep\Http\Middleware\Cache;
use EpTest\Http\ClientAwareTrait;
use GuzzleHttp\Psr7\Response;
use PHPUnit\Framework\TestCase;
use Psr\Cache\CacheItemPoolInterface;
use Psr\Http\Message\ResponseInterface;
use Symfony\Component\Cache\Adapter\ArrayAdapter;
use Zend\Diactoros\Response\HtmlResponse;

/**
 * Class CacheMiddlewareTest
 *
 * @package EpTest\Http\Middleware
 * @author Elektronika Perhoč <support@ep.hr>
 * @link https://bitbucket.org/elektronika-perhoc/ep-framework
 */
class CacheTest extends TestCase
{
    use ClientAwareTrait;

    /** @noinspection PhpDocMissingThrowsInspection */

    /**
     * @param int $default_lifetime
     *
     * @return CacheItemPoolInterface|ArrayAdapter
     */
    protected function createCache(int $default_lifetime = 0): CacheItemPoolInterface
    {
        $container = $this->createContainer();
        $factory = new AbstractCacheFactory();

        /** @noinspection PhpUnhandledExceptionInspection */
        return $factory($container, '', [
            'adapter' => 'array',
            'options' => [
                'default_lifetime' => $default_lifetime,
            ],
        ]);
    }

    protected function assertCacheHit(ResponseInterface $response)
    {
        $this->assertTrue($response->hasHeader(Cache::HEADER_CACHE_INFO));
        $this->assertEquals(Cache::HEADER_CACHE_HIT, $response->getHeaderLine(Cache::HEADER_CACHE_INFO));
    }

    protected function assertCacheMiss(ResponseInterface $response)
    {
        $this->assertTrue($response->hasHeader(Cache::HEADER_CACHE_INFO));
        $this->assertEquals(Cache::HEADER_CACHE_MISS, $response->getHeaderLine(Cache::HEADER_CACHE_INFO));
    }

    /**
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function testResponsesHaveCacheInfoHeaders()
    {
        $cache = $this->createCache();
        $client = $this->createClient([
            new HtmlResponse('1'),
            new HtmlResponse('1'),
        ], [
            'cache' => $cache
        ]);

        $response = $client->request('GET', '');
        $this->assertCacheMiss($response);

        $cachedResponse = $client->request('GET', '');
        $this->assertCacheHit($cachedResponse);

        $this->assertCount(1, array_filter($cache->getValues()));
    }

    /**
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function testErrorReponsesAreNotCached()
    {
        $cache = $this->createCache();
        $client = $this->createClient([
            new HtmlResponse('', 500),
            new HtmlResponse('', 500),
        ], [
            'cache' => $cache,
            'cache_lifetime' => 10,
        ]);

        $response = $client->request('GET', '');
        $this->assertCacheMiss($response);

        $cachedResponse = $client->request('GET', '');
        $this->assertCacheMiss($cachedResponse);

        $this->assertCount(0, array_filter($cache->getValues()));
    }

    /**
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function testResponseMaxAgeIsUsed()
    {
        $client = $this->createClient([
            new HtmlResponse('', 200, [
                'Cache-Control' => 'max-age=1',
            ]),
            new HtmlResponse('', 200, [
                'Cache-Control' => 'max-age=1',
            ]),
        ], [
            'cache' => $this->createCache(),
            'cache_lifetime' => -1,
        ]);

        $response = $client->request('GET', '');
        $this->assertCacheMiss($response);

        $cachedResponse = $client->request('GET', '');
        $this->assertCacheHit($cachedResponse);

        sleep(2);

        $staleResponse = $client->request('GET', '');
        $this->assertCacheMiss($staleResponse);
    }

    /**
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function testNotCachedIfMaxAgeIsZero()
    {
        $cache = $this->createCache();
        $client = $this->createClient([
            new HtmlResponse('', 200, [
                'Cache-Control' => 'max-age=0',
            ]),
            new HtmlResponse('', 200, [
                'Cache-Control' => 'max-age=0',
            ]),
        ], [
            'cache' => $cache,
            'cache_lifetime' => -1,
        ]);

        $response = $client->request('GET', '');
        $this->assertCacheMiss($response);

        $cachedResponse = $client->request('GET', '');
        $this->assertCacheMiss($cachedResponse);

        $this->assertCount(0, array_filter($cache->getValues()));
    }

    /**
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function testWillReturnProperResponseOnCacheMiss()
    {
        $client = $this->createClient([
            new Response(200, [
                'Content-Type' => 'application/json',
            ], '{"name":"value"}'),
        ], ['cache' => $this->createCache()]);

        $response = $client->request('GET', '');
        $content = $response->getBody()->getContents();

        $this->assertEquals('application/json', $response->getHeaderLine('Content-Type'));
        $this->assertNotEmpty($content);

        $decoded = json_decode($content, true);
        $this->assertEquals(['name' => 'value'], $decoded);
    }

    /**
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function testWillNotCacheIfNoCacheControlHeaders()
    {
        $cache = $this->createCache();
        $client = $this->createClient([
            new HtmlResponse('', 200, [
                'Cache-Control' => 'no-cache',
            ]),
            new HtmlResponse('', 200, [
                'Cache-Control' => 'no-cache',
            ]),
        ], ['cache' => $cache]);

        $response = $client->request('GET', '');
        $this->assertCacheMiss($response);

        $cachedResponse = $client->request('GET', '');
        $this->assertCacheMiss($cachedResponse);

        $this->assertCount(0, array_filter($cache->getValues()));
    }

    /**
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function testWillNotCacheIfDefaultLifetimeNegativeAndNoCacheControlHeader()
    {
        $cache = $this->createCache(-1);
        $client = $this->createClient([
            new HtmlResponse(''),
            new HtmlResponse(''),
        ], [
            'cache' => $cache,
            'cache_lifetime' => -1,
        ]);

        $response = $client->request('GET', '');
        $this->assertCacheMiss($response);

        $cachedResponse = $client->request('GET', '');
        $this->assertCacheMiss($cachedResponse);

        $this->assertCount(0, array_filter($cache->getValues()));
    }
}